#!/usr/bin/python
"""Class and helper functions for fitting the EVAC+ model."""

import logging

import numpy as np

from dipy.align.reslice import reslice
from dipy.data import get_fnames
from dipy.nn.utils import (
    normalize,
    recover_img,
    set_logger_level,
    transform_img,
)
from dipy.segment.utils import remove_holes_and_islands
from dipy.testing.decorators import doctest_skip_parser
from dipy.utils.deprecator import deprecated_params
from dipy.utils.optpkg import optional_package

torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
if have_torch:
    from torch.nn import (
        Conv3d,
        ConvTranspose3d,
        Dropout3d,
        LayerNorm,
        Module,
        ModuleList,
        ReLU,
        Softmax,
    )
else:

    class Module:
        pass

    logging.warning(
        "This model requires Pytorch.\
                    Please install these packages using \
                    pip."
    )

logging.basicConfig()
logger = logging.getLogger("EVAC+")


def prepare_img(image):
    """Function to prepare image for model input
    Specific to EVAC+

    Parameters
    ----------
    image : np.ndarray
        Input image

    Returns
    -------
    input_data : dict
    """
    input1 = np.moveaxis(image, -1, 0)
    input1 = np.expand_dims(input1, 1)

    input2, _ = reslice(image, np.eye(4), (1, 1, 1), (2, 2, 2))
    input2 = np.moveaxis(input2, -1, 0)
    input2 = np.expand_dims(input2, 1)

    input3, _ = reslice(image, np.eye(4), (1, 1, 1), (4, 4, 4))
    input3 = np.moveaxis(input3, -1, 0)
    input3 = np.expand_dims(input3, 1)

    input4, _ = reslice(image, np.eye(4), (1, 1, 1), (8, 8, 8))
    input4 = np.moveaxis(input4, -1, 0)
    input4 = np.expand_dims(input4, 1)

    input5, _ = reslice(image, np.eye(4), (1, 1, 1), (16, 16, 16))
    input5 = np.moveaxis(input5, -1, 0)
    input5 = np.expand_dims(input5, 1)

    input_data = [
        torch.from_numpy(input1).float(),
        torch.from_numpy(input2).float(),
        torch.from_numpy(input3).float(),
        torch.from_numpy(input4).float(),
        torch.from_numpy(input5).float(),
    ]

    return input_data


class MoveDimLayer(Module):
    def __init__(self, source_dim, dest_dim):
        super(MoveDimLayer, self).__init__()
        self.source_dim = source_dim
        self.dest_dim = dest_dim

    def forward(self, x):
        return torch.movedim(x, self.source_dim, self.dest_dim)


class ChannelSum(Module):
    def __init__(self):
        super(ChannelSum, self).__init__()

    def forward(self, inputs):
        return torch.sum(inputs, dim=1, keepdim=True)


class Add(Module):
    def __init__(self):
        super(Add, self).__init__()

    def forward(self, x, passed):
        return x + passed


class Block(Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        strides,
        padding,
        drop_r,
        n_layers,
        *,
        passed_channel=1,
        layer_type="down",
    ):
        super(Block, self).__init__()
        self.n_layers = n_layers
        self.layer_list = ModuleList()
        self.layer_list2 = ModuleList()

        cur_channel = in_channels
        for _ in range(n_layers):
            self.layer_list.append(
                Conv3d(
                    cur_channel,
                    out_channels,
                    kernel_size,
                    stride=strides,
                    padding=padding,
                )
            )
            cur_channel = out_channels
            self.layer_list.append(Dropout3d(drop_r))
            self.layer_list.append(MoveDimLayer(1, -1))
            self.layer_list.append(LayerNorm(out_channels))
            self.layer_list.append(MoveDimLayer(-1, 1))
            self.layer_list.append(ReLU())

        if layer_type == "down":
            self.layer_list2.append(Conv3d(in_channels, 1, 2, stride=2, padding=0))
            self.layer_list2.append(ReLU())
        elif layer_type == "up":
            self.layer_list2.append(
                ConvTranspose3d(passed_channel, 1, 2, stride=2, padding=0)
            )
            self.layer_list2.append(ReLU())

        self.channel_sum = ChannelSum()
        self.add = Add()

    def forward(self, input, passed):
        x = input
        for layer in self.layer_list:
            x = layer(x)

        x = self.channel_sum(x)
        fwd = self.add(x, passed)
        x = fwd

        for layer in self.layer_list2:
            x = layer(x)

        return fwd, x


class Model(Module):
    def __init__(self, model_scale=16):
        super(Model, self).__init__()

        # Block structure
        self.block1 = Block(
            1, model_scale, kernel_size=5, strides=1, padding=2, drop_r=0.2, n_layers=1
        )
        self.block2 = Block(
            2,
            model_scale * 2,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=2,
        )
        self.block3 = Block(
            2,
            model_scale * 4,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=3,
        )
        self.block4 = Block(
            2,
            model_scale * 8,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=3,
        )
        self.block5 = Block(
            2,
            model_scale * 16,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=3,
            passed_channel=2,
            layer_type="up",
        )

        # Upsample/decoder blocks
        self.up_block1 = Block(
            3,
            model_scale * 8,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=3,
            passed_channel=1,
            layer_type="up",
        )
        self.up_block2 = Block(
            3,
            model_scale * 4,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=3,
            passed_channel=1,
            layer_type="up",
        )
        self.up_block3 = Block(
            3,
            model_scale * 2,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=2,
            passed_channel=1,
            layer_type="up",
        )
        self.up_block4 = Block(
            2,
            model_scale,
            kernel_size=5,
            strides=1,
            padding=2,
            drop_r=0.5,
            n_layers=1,
            passed_channel=1,
            layer_type="none",
        )

        self.conv_pred = Conv3d(1, 2, 1, padding=0)
        self.softmax = Softmax(dim=1)

    def forward(self, inputs, raw_input_2, raw_input_3, raw_input_4, raw_input_5):
        fwd1, x = self.block1(inputs, inputs)
        x = torch.cat([x, raw_input_2], dim=1)

        fwd2, x = self.block2(x, x)
        x = torch.cat([x, raw_input_3], dim=1)

        fwd3, x = self.block3(x, x)
        x = torch.cat([x, raw_input_4], dim=1)

        fwd4, x = self.block4(x, x)
        x = torch.cat([x, raw_input_5], dim=1)

        # Decoding path
        _, up = self.block5(x, x)
        x = torch.cat([fwd4, up], dim=1)

        _, up = self.up_block1(x, up)
        x = torch.cat([fwd3, up], dim=1)

        _, up = self.up_block2(x, up)
        x = torch.cat([fwd2, up], dim=1)

        _, up = self.up_block3(x, up)
        x = torch.cat([fwd1, up], dim=1)

        _, pred = self.up_block4(x, up)

        pred = self.conv_pred(pred)
        output = self.softmax(pred)

        return output


class EVACPlus:
    """This class is intended for the EVAC+ model.

    The EVAC+ model :footcite:p:`Park2024` is a deep learning neural network for
    brain extraction. It uses a V-net architecture combined with
    multi-resolution input data, an additional conditional random field (CRF)
    recurrent layer and supplementary Dice loss term for this recurrent layer.

    References
    ----------
    .. footbibliography::
    """

    @doctest_skip_parser
    def __init__(self, *, verbose=False, use_cuda=False):
        """Model initialization

        The model was pre-trained for usage on
        brain extraction of T1 images.

        This model is designed to take as input
        a T1 weighted image.

        Parameters
        ----------
        verbose : bool, optional
            Whether to show information about the processing.
        use_cuda : bool, optional
            Whether to use GPU for processing.
            If False or no CUDA is detected, CPU will be used.
        """
        if not have_torch:
            raise torch()

        log_level = "INFO" if verbose else "CRITICAL"
        set_logger_level(log_level, logger)

        # EVAC+ network load

        self.model = self.init_model()
        if use_cuda and torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")
        self.model = self.model.to(self.device)
        self.fetch_default_weights()

    def init_model(self, model_scale=16):
        return Model(model_scale)

    def fetch_default_weights(self):
        """Load the model pre-training weights to use for the fitting.
        While the user can load different weights, the function
        is mainly intended for the class function 'predict'.
        """
        fetch_model_weights_path = get_fnames(name="evac_default_torch_weights")
        self.load_model_weights(fetch_model_weights_path)

    def load_model_weights(self, weights_path):
        """Load the custom pre-training weights to use for the fitting.

        Parameters
        ----------
        weights_path : str
            Path to the file containing the weights (pth, saved by Pytorch)
        """
        try:
            self.model.load_state_dict(
                torch.load(
                    weights_path,
                    weights_only=True,
                    map_location=self.device,
                )
            )
            self.model.eval()
        except ValueError as e:
            raise ValueError(
                "Expected input for the provided model weights \
                             do not match the declared model"
            ) from e

    def __predict(self, x_test):
        """Internal prediction function

        Parameters
        ----------
        x_test : list of np.ndarray
            Image should match the required shape of the model.

        Returns
        -------
        np.ndarray (batch, ...)
            Predicted brain mask
        """
        return self.model(*x_test)[:, 0].detach().numpy()

    @deprecated_params(
        "largest_area", new_name="finalize_mask", since="1.10", until="1.12"
    )
    def predict(
        self,
        T1,
        affine,
        *,
        voxsize=(1, 1, 1),
        batch_size=None,
        return_affine=False,
        return_prob=False,
        finalize_mask=True,
    ):
        """Wrapper function to facilitate prediction of larger dataset.

        Parameters
        ----------
        T1 : np.ndarray or list of np.ndarray
            For a single image, input should be a 3D array.
            If multiple images, it should be a a list or tuple.
        affine : np.ndarray (4, 4) or (batch, 4, 4)
            or list of np.ndarrays with len of batch
            Affine matrix for the T1 image. Should have
            batch dimension if T1 has one.
        voxsize : np.ndarray or list or tuple, optional
            (3,) or (batch, 3)
            voxel size of the T1 image.
        batch_size : int, optional
            Number of images per prediction pass. Only available if data
            is provided with a batch dimension.
            Consider lowering it if you get an out of memory error.
            Increase it if you want it to be faster and have a lot of data.
            If None, batch_size will be set to 1 if the provided image
            has a batch dimension.
        return_affine : bool, optional
            Whether to return the affine matrix. Useful if the input was a
            file path.
        return_prob : bool, optional
            Whether to return the probability map instead of a
            binary mask. Useful for testing.
        finalize_mask : bool, optional
            Whether to remove potential holes or islands.
            Useful for solving minor errors.

        Returns
        -------
        pred_output : np.ndarray (...) or (batch, ...)
            Predicted brain mask

        affine : np.ndarray (...) or (batch, ...)
            affine matrix of mask
            only if return_affine is True
        """
        voxsize = np.array(voxsize)
        affine = np.array(affine)

        if isinstance(T1, (list, tuple)):
            dim = 4
            T1 = np.array(T1)
        elif len(T1.shape) == 3:
            dim = 3
            if batch_size is not None:
                logger.warning(
                    "Batch size specified, but not used",
                    "due to the input not having \
                                a batch dimension",
                )

            T1 = np.expand_dims(T1, 0)
            affine = np.expand_dims(affine, 0)
            voxsize = np.expand_dims(voxsize, 0)
        else:
            raise ValueError(
                "T1 data should be a np.ndarray of dimension 3 or a list/tuple of it"
            )
        if batch_size is None:
            batch_size = 1

        input_data = np.zeros((128, 128, 128, len(T1)))
        affines = np.zeros((len(T1), 4, 4))
        mid_shapes = np.zeros((len(T1), 3)).astype(int)
        offset_arrays = np.zeros((len(T1), 4, 4)).astype(int)
        scales = np.zeros(len(T1))
        crop_vss = np.zeros((len(T1), 3, 2))
        pad_vss = np.zeros((len(T1), 3, 2))

        # Normalize the data.
        n_T1 = np.zeros(T1.shape)
        for i, T1_img in enumerate(T1):
            n_T1[i] = normalize(T1_img, new_min=0, new_max=1)
            t_img, t_affine, mid_shape, offset_array, scale, crop_vs, pad_vs = (
                transform_img(n_T1[i], affine[i], voxsize=voxsize[i])
            )
            input_data[..., i] = t_img
            affines[i] = t_affine
            mid_shapes[i] = mid_shape
            offset_arrays[i] = offset_array
            scales[i] = scale
            crop_vss[i] = crop_vs
            pad_vss[i] = pad_vs

        # Prediction stage
        prediction = np.zeros((len(T1), 128, 128, 128), dtype=np.float32)
        for batch_idx in range(batch_size, len(T1) + 1, batch_size):
            batch = input_data[..., batch_idx - batch_size : batch_idx]
            temp_input = prepare_img(batch)
            temp_pred = self.__predict(temp_input)
            prediction[:batch_idx] = temp_pred
        remainder = np.mod(len(T1), batch_size)
        if remainder != 0:
            temp_input = prepare_img(input_data[..., -remainder:])
            temp_pred = self.__predict(temp_input)
            prediction[-remainder:] = temp_pred

        output_mask = []
        for i in range(len(T1)):
            output = recover_img(
                prediction[i],
                affines[i],
                mid_shapes[i],
                n_T1[i].shape,
                offset_arrays[i],
                voxsize=voxsize[i],
                scale=scales[i],
                crop_vs=crop_vss[i],
                pad_vs=pad_vss[i],
            )
            if not return_prob:
                output = np.where(output >= 0.5, 1, 0)
                if finalize_mask:
                    output = remove_holes_and_islands(output, slice_wise=True)
            output_mask.append(output)

        if dim == 3:
            output_mask = output_mask[0]
            affine = affine[0]

        output_mask = np.array(output_mask)
        affine = np.array(affine)
        if return_affine:
            return output_mask, affine
        else:
            return output_mask



import numpy as np
from numpy.testing import assert_array_almost_equal, assert_equal, assert_raises

from dipy.align.streamlinear import (
    compose_matrix44,
    decompose_matrix44,
    slr_with_qbx,
    transform_streamlines,
    whole_brain_slr,
)
from dipy.data import get_fnames
from dipy.io.streamline import load_tractogram
from dipy.tracking.distances import bundles_distances_mam
from dipy.tracking.streamline import Streamlines


def test_whole_brain_slr():
    fname = get_fnames(name="fornix")

    fornix = load_tractogram(fname, "same", bbox_valid_check=False).streamlines

    f = Streamlines(fornix)
    f1 = f.copy()
    f2 = f.copy()

    # check translation
    f2._data += np.array([50, 0, 0])

    moved, transform, qb_centroids1, qb_centroids2 = whole_brain_slr(
        f1,
        f2,
        x0="affine",
        verbose=True,
        rm_small_clusters=2,
        greater_than=0,
        less_than=np.inf,
        qbx_thr=[5, 2, 1],
        progressive=False,
    )

    # we can check the quality of registration by comparing the matrices
    # MAM streamline distances before and after SLR
    D12 = bundles_distances_mam(f1, f2)
    D1M = bundles_distances_mam(f1, moved)

    d12_minsum = np.sum(np.min(D12, axis=0))
    d1m_minsum = np.sum(np.min(D1M, axis=0))

    print(f"distances= {d12_minsum} {d1m_minsum}")

    assert_equal(d1m_minsum < d12_minsum, True)

    assert_array_almost_equal(transform[:3, 3], [-50, -0, -0], 2)

    # check rotation

    mat = compose_matrix44([0, 0, 0, 15, 0, 0])

    f3 = f.copy()
    f3 = transform_streamlines(f3, mat)

    moved, transform, qb_centroids1, qb_centroids2 = slr_with_qbx(
        f1,
        f3,
        verbose=False,
        rm_small_clusters=1,
        greater_than=20,
        less_than=np.inf,
        qbx_thr=[2],
        progressive=True,
    )

    # we can also check the quality by looking at the decomposed transform

    assert_array_almost_equal(decompose_matrix44(transform)[3], -15, 2)

    moved, transform, qb_centroids1, qb_centroids2 = slr_with_qbx(
        f1,
        f3,
        verbose=False,
        rm_small_clusters=1,
        select_random=400,
        greater_than=20,
        less_than=np.inf,
        qbx_thr=[2],
        progressive=True,
    )

    # we can also check the quality by looking at the decomposed transform

    assert_array_almost_equal(decompose_matrix44(transform)[3], -15, 2)


def test_slr_one_streamline():
    fname = get_fnames(name="fornix")

    fornix = load_tractogram(fname, "same", bbox_valid_check=False).streamlines

    f = Streamlines(fornix)
    f1_one = Streamlines([f[0]])
    f2 = f.copy()
    f2._data += np.array([50, 0, 0])

    assert_raises(
        ValueError,
        slr_with_qbx,
        f1_one,
        f2,
        verbose=False,
        rm_small_clusters=50,
        greater_than=20,
        less_than=np.inf,
        qbx_thr=[2],
        progressive=True,
    )

    assert_raises(
        ValueError,
        slr_with_qbx,
        f2,
        f1_one,
        verbose=False,
        rm_small_clusters=50,
        greater_than=20,
        less_than=np.inf,
        qbx_thr=[2],
        progressive=True,
    )



"""
This file is used by asv_compare.conf.json.tpl.

Note
----

This file is copied (possibly with major modifications) from the
sources of the numpy project - https://github.com/numpy/numpy.
It remains licensed as the rest of NUMPY (BSD 3-Clause as of November 2023).

# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the Numpy package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""

import subprocess
import sys

# pip ignores '--global-option' when pep517 is enabled therefore we disable it.
cmd = [sys.executable, "-mpip", "wheel", "--no-use-pep517"]
try:
    output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, text=True)
except Exception as e:
    output = str(e.output)
if "no such option" in output:
    print("old version of pip, escape '--no-use-pep517'")
    cmd.pop()

subprocess.run(cmd + sys.argv[1:])



from functools import partial
from pathlib import Path
import warnings

import numpy as np

from dipy.testing.decorators import is_macOS, warning_for_keywords
from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.tab import (
    HorizonTab,
    build_checkbox,
    build_label,
    build_slider,
    build_switcher,
)

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from fury import colormap


class SlicesTab(HorizonTab):
    """Interaction tab for slice visualization.

    Attributes
    ----------
    name : str
        Name of the tab.
    """

    @warning_for_keywords()
    def __init__(
        self,
        slices_visualizer,
        tab_name,
        file_name,
        *,
        force_render=lambda _element: None,
    ):
        super().__init__()

        self._visualizer = slices_visualizer
        self._name = tab_name
        self._file_name = Path(file_name or tab_name).name
        self._force_render = force_render
        self._tab_id = 0

        self.on_slice_change = lambda _tab_id, _x, _y, _z: None
        self.on_volume_change = lambda _tab_id, v: None

        self._actor_toggle = build_checkbox(
            labels=[""], checked_labels=[""], on_change=self._toggle_actors
        )

        self._slice_opacity_label, self._slice_opacity = build_slider(
            initial_value=1.0,
            max_value=1.0,
            text_template="{ratio:.0%}",
            on_change=self._change_opacity,
            label="Opacity",
        )

        self._slice_x_label, self._slice_x = build_slider(
            initial_value=self._visualizer.selected_slices[0],
            max_value=self._visualizer.data_shape[0] - 1,
            text_template="{value:.0f}",
            label="X Slice",
        )

        self._change_slice_x = partial(
            self._change_slice_value, selected_slice=self._slice_x
        )
        self._adjust_slice_x = partial(
            self._change_slice_value, selected_slice=self._slice_x, sync_slice=True
        )

        self._slice_x.obj.on_value_changed = self._adjust_slice_x
        self._slice_x.obj.on_moving_slider = self._change_slice_x

        self._change_slice_visibility_x = partial(
            self._update_slice_visibility, selected_slice=self._slice_x, actor_idx=0
        )

        self._slice_x_toggle = build_checkbox(
            labels=[""], checked_labels=[""], on_change=self._change_slice_visibility_x
        )

        self._slice_y_label, self._slice_y = build_slider(
            initial_value=self._visualizer.selected_slices[1],
            max_value=self._visualizer.data_shape[1] - 1,
            text_template="{value:.0f}",
            label="Y Slice",
        )

        self._change_slice_y = partial(
            self._change_slice_value, selected_slice=self._slice_y
        )
        self._adjust_slice_y = partial(
            self._change_slice_value, selected_slice=self._slice_y, sync_slice=True
        )

        self._slice_y.obj.on_value_changed = self._adjust_slice_y
        self._slice_y.obj.on_moving_slider = self._change_slice_y

        self._change_slice_visibility_y = partial(
            self._update_slice_visibility, selected_slice=self._slice_y, actor_idx=1
        )

        self._slice_y_toggle = build_checkbox(
            labels=[""], checked_labels=[""], on_change=self._change_slice_visibility_y
        )

        self._slice_z_label, self._slice_z = build_slider(
            initial_value=self._visualizer.selected_slices[2],
            max_value=self._visualizer.data_shape[2] - 1,
            text_template="{value:.0f}",
            label="Z Slice",
        )

        self._change_slice_z = partial(
            self._change_slice_value, selected_slice=self._slice_z
        )
        self._adjust_slice_z = partial(
            self._change_slice_value, selected_slice=self._slice_z, sync_slice=True
        )

        self._slice_z.obj.on_value_changed = self._adjust_slice_z
        self._slice_z.obj.on_moving_slider = self._change_slice_z

        self._change_slice_visibility_z = partial(
            self._update_slice_visibility, selected_slice=self._slice_z, actor_idx=2
        )

        self._slice_z_toggle = build_checkbox(
            labels=[""], checked_labels=[""], on_change=self._change_slice_visibility_z
        )

        self._voxel_label = build_label(text="Voxel")
        self._voxel_data = build_label(text="")

        self._visualizer.register_picker_callback(self._change_picked_voxel)

        self._register_visualize_partials()

        self._file_label = build_label(text="Filename")
        self._file_name_label = build_label(text=self._file_name)

        self._register_elements(
            self._actor_toggle,
            self._slice_opacity_label,
            self._slice_opacity,
            self._slice_x_toggle,
            self._slice_x_label,
            self._slice_x,
            self._slice_y_toggle,
            self._slice_y_label,
            self._slice_y,
            self._slice_z_toggle,
            self._slice_z_label,
            self._slice_z,
            self._voxel_label,
            self._voxel_data,
            self._file_label,
            self._file_name_label,
        )

        if not self._visualizer.rgb:
            self._intensities_label, self._intensities = build_slider(
                initial_value=self._visualizer.intensities_range,
                min_value=self._visualizer.volume_min,
                max_value=self._visualizer.volume_max,
                text_template="{value:.0f}",
                on_change=self._change_intensity,
                label="Intensities",
                is_double_slider=True,
            )

            self._supported_colormap = [
                {"label": "Gray", "value": "gray"},
                {"label": "Bone", "value": "bone"},
                {"label": "Cividis", "value": "cividis"},
                {"label": "Inferno", "value": "inferno"},
                {"label": "Magma", "value": "magma"},
                {"label": "Viridis", "value": "viridis"},
                {"label": "Jet", "value": "jet"},
                {"label": "Pastel 1", "value": "Pastel1"},
                {"label": "Distinct", "value": "dist"},
            ]

            self._colormap_switcher_label, self._colormap_switcher = build_switcher(
                items=self._supported_colormap,
                label="Colormap",
                on_value_changed=self._change_color_map,
            )

            self._register_elements(
                self._intensities_label,
                self._intensities,
                self._colormap_switcher_label,
                self._colormap_switcher,
            )

            if len(self._visualizer.data_shape) == 4:
                self._adjust_volume = partial(self._change_volume, sync_vol=True)
                self._volume_label, self._volume = build_slider(
                    initial_value=0,
                    max_value=self._visualizer.data_shape[-1] - 1,
                    on_moving_slider=self._change_volume,
                    on_value_changed=self._adjust_volume,
                    text_template="{value:.0f}",
                    label="Volume",
                )
                self._register_elements(self._volume_label, self._volume)

    def _register_visualize_partials(self):
        self._visualize_slice_x = partial(
            self._visualizer.slice_actors[0].display_extent,
            y1=0,
            y2=self._visualizer.data_shape[1] - 1,
            z1=0,
            z2=self._visualizer.data_shape[2] - 1,
        )
        self._visualize_slice_y = partial(
            self._visualizer.slice_actors[1].display_extent,
            x1=0,
            x2=self._visualizer.data_shape[0] - 1,
            z1=0,
            z2=self._visualizer.data_shape[2] - 1,
        )
        self._visualize_slice_z = partial(
            self._visualizer.slice_actors[2].display_extent,
            x1=0,
            x2=self._visualizer.data_shape[0] - 1,
            y1=0,
            y2=self._visualizer.data_shape[1] - 1,
        )

    def _change_color_map(self, _idx, _value):
        self._update_colormap()
        self._force_render(self)

    def _change_intensity(self, slider):
        self._intensities.selected_value[0] = slider.left_disk_value
        self._intensities.selected_value[1] = slider.right_disk_value
        self._update_colormap()

    def _change_opacity(self, slider):
        self._slice_opacity.selected_value = slider.value
        self._update_opacities()

    def _change_picked_voxel(self, message):
        self._voxel_data.obj.message = message
        self._voxel_data.selected_value = message

    @warning_for_keywords()
    def _change_slice_value(self, slider, selected_slice, *, sync_slice=False):
        selected_slice.selected_value = int(np.rint(slider.value))
        if not sync_slice:
            self.on_slice_change(
                self._tab_id,
                self._slice_x.selected_value,
                self._slice_y.selected_value,
                self._slice_z.selected_value,
            )
        self._visualize_slice_x(
            x1=self._slice_x.selected_value, x2=self._slice_x.selected_value
        )
        self._visualize_slice_y(
            y1=self._slice_y.selected_value, y2=self._slice_y.selected_value
        )
        self._visualize_slice_z(
            z1=self._slice_z.selected_value, z2=self._slice_z.selected_value
        )

    @warning_for_keywords()
    def _update_slice_visibility(
        self, checkboxes, selected_slice, actor_idx, *, visibility=None
    ):
        if checkboxes is not None and "" in checkboxes.checked_labels:
            visibility = True
        elif visibility is None:
            visibility = False

        selected_slice.visibility = visibility
        selected_slice.obj.set_visibility(visibility)
        self._visualizer.slice_actors[actor_idx].SetVisibility(visibility)

    def _change_volume(self, slider, sync_vol=False):
        value = int(np.rint(slider.value))
        if value != self._volume.selected_value:
            if not sync_vol:
                self.on_volume_change(self._tab_id, value)
            visible_slices = (
                self._slice_x.selected_value,
                self._slice_y.selected_value,
                self._slice_z.selected_value,
            )
            valid_vol, default_range = self._visualizer.change_volume(
                value,
                np.asarray(self._intensities.obj._ratio),
                visible_slices,
            )
            self._register_visualize_partials()
            if not valid_vol:
                warnings.warn(
                    f"Volume N°{value} does not have any contrast. Please, "
                    "check the value ranges of your data. Returning to "
                    "previous volume.",
                    stacklevel=2,
                )
                self._volume.obj.value = self._volume.selected_value
            else:
                intensities_range = self._visualizer.intensities_range

                # Updating the colormap
                self._intensities.selected_value[0] = intensities_range[0]
                self._intensities.selected_value[1] = intensities_range[1]
                self._update_colormap()

                # Updating intensities slider
                if default_range:
                    self._intensities.obj.left_disk_value = self._visualizer.volume_min
                    self._intensities.obj.right_disk_value = self._visualizer.volume_max
                self._intensities.obj.min_value = self._visualizer.volume_min
                self._intensities.obj.max_value = self._visualizer.volume_max
                self._intensities.obj.update(0)
                self._intensities.obj.update(1)

                # Updating opacities
                self._update_opacities()

                # Updating visibilities
                slices = [self._slice_x, self._slice_y, self._slice_z]
                for i, s in enumerate(slices):
                    self._update_slice_visibility(None, s, i, visibility=s.visibility)

                self._volume.selected_value = value
                self._force_render(self)

    def _update_colormap(self):
        if self._colormap_switcher.selected_value[1] == "dist":
            rgb = colormap.distinguishable_colormap(nb_colors=256)
            rgb = np.asarray(rgb)
        else:
            rgb = colormap.create_colormap(
                np.linspace(
                    self._intensities.selected_value[0],
                    self._intensities.selected_value[1],
                    256,
                ),
                name=self._colormap_switcher.selected_value[1],
                auto=True,
            )
        num_lut = rgb.shape[0]

        lut = colormap.LookupTable()
        lut.SetNumberOfTableValues(num_lut)
        lut.SetRange(
            self._intensities.selected_value[0], self._intensities.selected_value[1]
        )
        for i in range(num_lut):
            r, g, b = rgb[i]
            lut.SetTableValue(i, r, g, b)
        lut.SetRampToLinear()
        lut.Build()

        for slice_actor in self._visualizer.slice_actors:
            slice_actor.output.SetLookupTable(lut)
            slice_actor.output.Update()

    def _update_opacities(self):
        for slice_actor in self._visualizer.slice_actors:
            slice_actor.GetProperty().SetOpacity(self._slice_opacity.selected_value)

    def on_tab_selected(self):
        """Trigger when tab becomes active."""
        super().on_tab_selected()
        self._slice_x.obj.set_visibility(self._slice_x.visibility)
        self._slice_y.obj.set_visibility(self._slice_y.visibility)
        self._slice_z.obj.set_visibility(self._slice_z.visibility)

    def update_slices(self, x_slice, y_slice, z_slice):
        """Updates slicer positions.

        Parameters
        ----------
        x_slice: float
            x-value where the slicer should be placed
        y_slice: float
            y-value where the slicer should be placed
        z_slice: float
            z-value where the slicer should be placed
        """
        if not self._slice_x.obj.value == x_slice:
            self._slice_x.obj.value = x_slice

        if not self._slice_y.obj.value == y_slice:
            self._slice_y.obj.value = y_slice

        if not self._slice_z.obj.value == z_slice:
            self._slice_z.obj.value = z_slice

    def update_volume(self, volume):
        """Updates volume based on passed volume.

        Parameters
        ----------
        volume : float
            value of where the volume slider should be placed
        """

        if not hasattr(self, "_volume"):
            return

        if not self._volume.obj.value == volume:
            self._volume.obj.value = volume

    def build(self, tab_id):
        """Build all the elements under the tab.

        Parameters
        ----------
        tab_id : int
            Id of the tab.
        """
        self._tab_id = tab_id

        x_pos = 0.02
        self._actor_toggle.position = (x_pos, 0.85)
        self._slice_x_toggle.position = (x_pos, 0.62)
        self._slice_y_toggle.position = (x_pos, 0.38)
        self._slice_z_toggle.position = (x_pos, 0.15)

        x_pos = 0.05
        self._slice_opacity_label.position = (x_pos, 0.85)
        self._slice_x_label.position = (x_pos, 0.62)
        self._slice_y_label.position = (x_pos, 0.38)
        self._slice_z_label.position = (x_pos, 0.15)

        x_pos = 0.10
        self._slice_opacity.position = (x_pos, 0.85)
        self._slice_x.position = (x_pos, 0.62)
        self._slice_y.position = (x_pos, 0.38)
        self._slice_z.position = (x_pos, 0.15)

        x_pos = 0.52
        self._voxel_label.position = (x_pos, 0.42)
        self._file_label.position = (x_pos, 0.28)

        x_pos = 0.60
        self._voxel_data.position = (x_pos, 0.42)
        self._file_name_label.position = (x_pos, 0.28)

        if is_macOS:
            self._file_name_label.size = (800, "auto")
        else:
            self._file_name_label.size = (400, "auto")

        if not self._visualizer.rgb:
            x_pos = 0.52
            self._intensities_label.position = (x_pos, 0.85)
            self._colormap_switcher_label.position = (x_pos, 0.56)

            x_pos = 0.60
            self._intensities.position = (x_pos, 0.85)
            self._colormap_switcher.position = [
                (x_pos, 0.54),
                (0.63, 0.54),
                (0.69, 0.54),
            ]

            if len(self._visualizer.data_shape) == 4:
                x_pos = 0.52
                self._volume_label.position = (x_pos, 0.15)

                x_pos = 0.60
                self._volume.position = (x_pos, 0.15)

    @property
    def name(self):
        """Name of the tab."""
        return self._name

    @property
    def file_name(self):
        """Name of the file opened in the tab."""
        return self._file_name

    @property
    def actors(self):
        """visualization actors controlled by tab."""
        return self._visualizer.slice_actors



import numpy as np
from numpy.testing import assert_, assert_array_almost_equal, assert_raises

from dipy.denoise.gibbs import (
    _gibbs_removal_1d,
    _gibbs_removal_2d,
    _image_tv,
    gibbs_removal,
)


def setup_module():
    """Module-level setup"""
    global image_gibbs, image_gt, image_cor, Nre

    # Produce a 2D image
    Nori = 32
    image = np.zeros((6 * Nori, 6 * Nori))
    image[Nori : 2 * Nori, Nori : 2 * Nori] = 1
    image[Nori : 2 * Nori, 4 * Nori : 5 * Nori] = 1
    image[2 * Nori : 3 * Nori, Nori : 3 * Nori] = 1
    image[3 * Nori : 4 * Nori, 2 * Nori : 3 * Nori] = 2
    image[3 * Nori : 4 * Nori, 4 * Nori : 5 * Nori] = 1
    image[4 * Nori : 5 * Nori, 3 * Nori : 5 * Nori] = 3

    # Corrupt image with gibbs ringing
    c = np.fft.fft2(image)
    c = np.fft.fftshift(c)
    c_crop = c[48:144, 48:144]
    image_gibbs = abs(np.fft.ifft2(c_crop) / 4)

    # Produce ground truth
    Nre = 16
    image_gt = np.zeros((6 * Nre, 6 * Nre))
    image_gt[Nre : 2 * Nre, Nre : 2 * Nre] = 1
    image_gt[Nre : 2 * Nre, 4 * Nre : 5 * Nre] = 1
    image_gt[2 * Nre : 3 * Nre, Nre : 3 * Nre] = 1
    image_gt[3 * Nre : 4 * Nre, 2 * Nre : 3 * Nre] = 2
    image_gt[3 * Nre : 4 * Nre, 4 * Nre : 5 * Nre] = 1
    image_gt[4 * Nre : 5 * Nre, 3 * Nre : 5 * Nre] = 3

    # Suppressing gibbs artefacts
    image_cor = _gibbs_removal_2d(image_gibbs)


def test_parallel():
    # Only relevant for 3d or 4d inputs

    # Make input data
    input_2d = image_gibbs.copy()
    input_3d = np.stack([input_2d, input_2d], axis=2)
    input_4d = np.stack([input_3d, input_3d], axis=3)

    # Test 3d case
    output_3d_parallel = gibbs_removal(input_3d, inplace=False, num_processes=2)
    output_3d_no_parallel = gibbs_removal(input_3d, inplace=False, num_processes=1)
    assert_array_almost_equal(output_3d_parallel, output_3d_no_parallel)

    # Test 4d case
    output_4d_parallel = gibbs_removal(input_4d, inplace=False, num_processes=2)
    output_4d_no_parallel = gibbs_removal(input_4d, inplace=False, num_processes=1)
    assert_array_almost_equal(output_4d_parallel, output_4d_no_parallel)

    # Test num_processes=None case
    output_4d_all_cpu = gibbs_removal(input_4d, inplace=False, num_processes=None)
    assert_array_almost_equal(output_4d_all_cpu, output_4d_no_parallel)


def test_inplace():
    # Make input data
    input_2d = image_gibbs.copy()
    input_3d = np.stack([input_2d, input_2d], axis=2)
    input_4d = np.stack([input_3d, input_3d], axis=3)

    # Test 2d cases
    output_2d = gibbs_removal(input_2d, inplace=False)
    assert_raises(AssertionError, assert_array_almost_equal, input_2d, output_2d)

    output_2d = gibbs_removal(input_2d, inplace=True)
    assert_array_almost_equal(input_2d, output_2d)

    # Test 3d case
    output_3d = gibbs_removal(input_3d, inplace=False)
    assert_raises(AssertionError, assert_array_almost_equal, input_3d, output_3d)

    output_3d = gibbs_removal(input_3d, inplace=True)
    assert_array_almost_equal(input_3d, output_3d)

    # Test 4d case
    output_4d = gibbs_removal(input_4d, inplace=False)
    assert_raises(AssertionError, assert_array_almost_equal, input_4d, output_4d)

    output_4d = gibbs_removal(input_4d, inplace=True)
    assert_array_almost_equal(input_4d, output_4d)


def test_gibbs_2d():
    # Correction of gibbs ringing have to be closer to gt than denoised image
    diff_raw = np.mean(abs(image_gibbs - image_gt))
    diff_cor = np.mean(abs(image_cor - image_gt))
    assert_(diff_raw > diff_cor)

    # Test if gibbs_removal works for 2D data
    image_cor2 = gibbs_removal(image_gibbs, inplace=False)
    assert_array_almost_equal(image_cor2, image_cor)


def test_gibbs_3d():
    image3d = np.zeros((6 * Nre, 6 * Nre, 2))
    image3d[:, :, 0] = image_gibbs
    image3d[:, :, 1] = image_gibbs

    image3d_cor = gibbs_removal(image3d, slice_axis=2)
    assert_array_almost_equal(image3d_cor[:, :, 0], image_cor)
    assert_array_almost_equal(image3d_cor[:, :, 1], image_cor)


def test_gibbs_4d():
    image4d = np.zeros((6 * Nre, 6 * Nre, 2, 2))
    image4d[:, :, 0, 0] = image_gibbs
    image4d[:, :, 1, 0] = image_gibbs
    image4d[:, :, 0, 1] = image_gibbs
    image4d[:, :, 1, 1] = image_gibbs

    image4d_cor = gibbs_removal(image4d)
    assert_array_almost_equal(image4d_cor[:, :, 0, 0], image_cor)
    assert_array_almost_equal(image4d_cor[:, :, 1, 0], image_cor)
    assert_array_almost_equal(image4d_cor[:, :, 0, 1], image_cor)
    assert_array_almost_equal(image4d_cor[:, :, 1, 1], image_cor)


def test_swapped_gibbs_2d():
    # 2D case: In this case slice_axis is a dummy variable. Since data is
    # already a single 2D image, to axis swapping is required
    image_cor0 = gibbs_removal(image_gibbs, slice_axis=0, inplace=False)
    assert_array_almost_equal(image_cor0, image_cor)

    image_cor1 = gibbs_removal(image_gibbs, slice_axis=1, inplace=False)
    assert_array_almost_equal(image_cor1, image_cor)

    image_cor2 = gibbs_removal(image_gibbs, slice_axis=2, inplace=False)
    assert_array_almost_equal(image_cor2, image_cor)


def test_swapped_gibbs_3d():
    image3d = np.zeros((6 * Nre, 2, 6 * Nre))
    image3d[:, 0, :] = image_gibbs
    image3d[:, 1, :] = image_gibbs

    image3d_cor = gibbs_removal(image3d, slice_axis=1)
    assert_array_almost_equal(image3d_cor[:, 0, :], image_cor)
    assert_array_almost_equal(image3d_cor[:, 1, :], image_cor)

    image3d = np.zeros((2, 6 * Nre, 6 * Nre))
    image3d[0, :, :] = image_gibbs
    image3d[1, :, :] = image_gibbs

    image3d_cor = gibbs_removal(image3d, slice_axis=0)
    assert_array_almost_equal(image3d_cor[0, :, :], image_cor)
    assert_array_almost_equal(image3d_cor[1, :, :], image_cor)


def test_swapped_gibbs_4d():
    image4d = np.zeros((2, 6 * Nre, 6 * Nre, 2))
    image4d[0, :, :, 0] = image_gibbs
    image4d[1, :, :, 0] = image_gibbs
    image4d[0, :, :, 1] = image_gibbs
    image4d[1, :, :, 1] = image_gibbs

    image4d_cor = gibbs_removal(image4d, slice_axis=0)
    assert_array_almost_equal(image4d_cor[0, :, :, 0], image_cor)
    assert_array_almost_equal(image4d_cor[1, :, :, 0], image_cor)
    assert_array_almost_equal(image4d_cor[0, :, :, 1], image_cor)
    assert_array_almost_equal(image4d_cor[1, :, :, 1], image_cor)


def test_gibbs_errors():
    assert_raises(ValueError, gibbs_removal, np.ones((2, 2, 2, 2, 2)))
    assert_raises(ValueError, gibbs_removal, np.ones(2))
    assert_raises(ValueError, gibbs_removal, np.ones((2, 2, 2)), slice_axis=3)
    assert_raises(TypeError, gibbs_removal, image_gibbs.copy(), inplace="True")
    # Test for valid num_processes
    assert_raises(TypeError, gibbs_removal, image_gibbs.copy(), num_processes="1")
    assert_raises(ValueError, gibbs_removal, image_gibbs.copy(), num_processes=0)
    # Test for valid input dimensionality
    assert_raises(ValueError, gibbs_removal, np.ones(2))  # 1D
    assert_raises(ValueError, gibbs_removal, np.ones((2, 2, 2, 2, 2)))  # 5D


def test_gibbs_subfunction():
    # This complementary test is to make sure that Gibbs suppression
    # sub-functions are properly implemented

    # Testing correction along axis 0
    image_a0 = _gibbs_removal_1d(image_gibbs, axis=0)
    # After this step tv along axis 0 should provide lower values than along
    # axis 1
    tv0_a0_r, tv0_a0_l = _image_tv(image_a0, axis=0)
    tv0_a0 = np.minimum(tv0_a0_r, tv0_a0_l)
    tv1_a0_r, tv1_a0_l = _image_tv(image_a0, axis=1)
    tv1_a0 = np.minimum(tv1_a0_r, tv1_a0_l)
    # Let's check that
    mean_tv0 = np.mean(abs(tv0_a0))
    mean_tv1 = np.mean(abs(tv1_a0))
    assert_(mean_tv0 < mean_tv1)

    # Testing correction along axis 1
    image_a1 = _gibbs_removal_1d(image_gibbs, axis=1)
    # After this step tv along axis 1 should provide higher values than along
    # axis 0
    tv0_a1_r, tv0_a1_l = _image_tv(image_a1, axis=0)
    tv0_a1 = np.minimum(tv0_a1_r, tv0_a1_l)
    tv1_a1_r, tv1_a1_l = _image_tv(image_a1, axis=1)
    tv1_a1 = np.minimum(tv1_a1_r, tv1_a1_l)
    # Let's check that
    mean_tv0 = np.mean(abs(tv0_a1))
    mean_tv1 = np.mean(abs(tv1_a1))
    assert_(mean_tv0 > mean_tv1)


def test_non_square_image():
    # Produce non-square 2D image
    Nori = 32
    img = np.zeros((6 * Nori, 6 * Nori))
    img[Nori : 2 * Nori, Nori : 2 * Nori] = 1
    img[2 * Nori : 3 * Nori, Nori : 3 * Nori] = 1
    img[3 * Nori : 4 * Nori, 2 * Nori : 3 * Nori] = 2
    img[4 * Nori : 5 * Nori, 3 * Nori : 5 * Nori] = 3

    # Corrupt image with gibbs ringing
    c = np.fft.fft2(img)
    c = np.fft.fftshift(c)
    c_crop = c[48:144, :]
    img_gibbs = abs(np.fft.ifft2(c_crop) / 2)

    # Produce ground truth
    Nre = 16
    img_gt = np.zeros((6 * Nre, 6 * Nori))
    img_gt[Nre : 2 * Nre, Nori : 2 * Nori] = 1
    img_gt[2 * Nre : 3 * Nre, Nori : 3 * Nori] = 1
    img_gt[3 * Nre : 4 * Nre, 2 * Nori : 3 * Nori] = 2
    img_gt[4 * Nre : 5 * Nre, 3 * Nori : 5 * Nori] = 3

    # Suppressing gibbs artefacts
    img_cor = gibbs_removal(img_gibbs, inplace=False)

    # Correction of gibbs ringing have to be closer to gt than denoised image
    diff_raw = np.mean(abs(img_gibbs - img_gt))
    diff_cor = np.mean(abs(img_cor - img_gt))
    assert_(diff_raw > diff_cor)
